"""
This module contains the c++ shared object files which are accessed via python.
"""
from __future__ import absolute_import
try:
	try:
		from . import libnecsim
	except ImportError:
		import libnecsim
	NECSimError = libnecsim.NECSimError
except ImportError as ie:
	import logging
	logging.warning("Cannot import libnecsim: {}".format(ie))
	libnecsim = None
	NECSimError = None



"""
Combine simulation outputs from separate guilds. Detailed :ref:`here <merging_simulations>`.

:py:class:`~Merger` will output a single database file, merging the various biodiversity tables into one.

Metrics are also calculated for the entire system, with a guild reference of 0.

All standard routines provided in :py:class:`~pycoalescence.coal_analyse.CoalescenceTree` can then be performed on the
combined database.
"""
from __future__ import absolute_import

import logging
import os

try:
	import sqlite3
except ImportError:
	# Python 3 compatibility
	import sqlite as sqlite3
try:
	from math import isclose
except ImportError:
	from .system_operations import isclose

from .coalescence_tree import CoalescenceTree
from .sqlite_connection import check_sql_table_exist, fetch_table_from_sql
from .system_operations import cantor_pairing, check_parent


class Merger(CoalescenceTree):
	"""
	Merges simulation outputs into a single database. Inherits from
	:py:class:`~pycoalescence.coal_analyse.CoalescenceTree` to provide all routines in the same object.
	"""
	def __init__(self, database=None, logging_level=logging.WARNING, log_output=None):
		super(Merger, self).__init__(database, logging_level, log_output)
		self.species_list_parent_counter = 0
		self.parameters_list = []
		self.guild = 0
		self.simulation_list = []
		self.species_list = []
		self.species_locations = []
		self.species_abundances = []
		self.fragment_abundances = []
		self.species_richness = []
		self.fragment_richness = []
		self.fragment_octaves = []
		self.community_parameters = []
		self.metacommunity_parameters = []
		self.max_species_id = 0
		self.max_locations_id = 0

	def __del__(self):
		"""
		Closes the connection to the database.
		"""
		if self.database:
			self.database.close()
			self.database = None

	def set_database(self, filename):
		"""
		Sets the output database for the merged simulations

		Assumes no database currently exists, and will create one.

		:raises IOError: if the output database already exists

		:param filename: the filename to output merged simulations into

		:rtype: None
		"""
		if os.path.exists(filename):
			raise IOError("Database already exists at {}".format(filename))
		self.file = filename
		check_parent(filename)
		try:
			self.database = sqlite3.connect(filename)
		except sqlite3.OperationalError as e:
			try:
				self.database.close()
			except AttributeError:
				pass
			self.database = None
			raise IOError("Error opening SQLite database: " + str(e))
		self._create_simulation_parameters()
		self._create_species_list()

	def _create_simulation_parameters(self):
		"""
		Creates the SIMULATION_PARAMETERS table in the database.

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a SIMULATION_PARAMETERS table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE SIMULATION_PARAMETERS (seed INT not null, job_type INT NOT NULL, " \
					 "output_dir TEXT NOT NULL, speciation_rate DOUBLE NOT NULL, sigma DOUBLE NOT NULL,tau DOUBLE NOT " \
					 "NULL, deme INT NOT NULL, sample_size DOUBLE NOT NULL, max_time INT NOT NULL, " \
					 "dispersal_relative_cost DOUBLE NOT NULL, min_num_species INT NOT NULL, " \
					 "habitat_change_rate DOUBLE NOT NULL, gen_since_historical DOUBLE NOT NULL, " \
					 "time_config_file TEXT NOT NULL, coarse_map_file TEXT NOT NULL, coarse_map_x INT NOT NULL, " \
					 "coarse_map_y INT NOT NULL, coarse_map_x_offset INT NOT NULL, coarse_map_y_offset INT NOT NULL, " \
					 "coarse_map_scale DOUBLE NOT NULL, fine_map_file TEXT NOT NULL, fine_map_x INT NOT NULL, " \
					 "fine_map_y INT NOT NULL, fine_map_x_offset INT NOT NULL, fine_map_y_offset INT NOT NULL, " \
					 "sample_file TEXT NOT NULL, grid_x INT NOT NULL, grid_y INT NOT NULL, sample_x INT NOT NULL, " \
					 "sample_y INT NOT NULL, sample_x_offset INT NOT NULL, sample_y_offset INT NOT NULL, " \
					 "historical_coarse_map TEXT NOT NULL, historical_fine_map TEXT NOT NULL, sim_complete INT NOT NULL, " \
					 "dispersal_method TEXT NOT NULL, m_probability DOUBLE NOT NULL, cutoff DOUBLE NOT NULL, " \
					 "restrict_self INT NOT NULL, landscape_type TEXT NOT NULL, protracted INT NOT NULL, " \
					 "min_speciation_gen DOUBLE NOT NULL, max_speciation_gen DOUBLE NOT NULL, " \
					 "dispersal_map TEXT NOT NULL, guild INT PRIMARY KEY NOT NULL, filename TEXT NOT NULL);"
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not create SIMULATION_PARAMETERS table in database {}: {}".format(self.file, soe))

	def _create_community_parameters(self):
		"""
		Creates the COMMUNITY_PARAMETERS table in the database.

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a COMMUNITY_PARAMETERS table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS COMMUNITY_PARAMETERS (reference INT PRIMARY KEY NOT NULL," \
					 " speciation_rate DOUBLE NOT NULL, time DOUBLE NOT NULL, fragments INT NOT NULL, " \
					 "metacommunity_reference INT);"
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not create COMMUNITY_PARAMETERS table in database {}: {}".format(self.file, soe))

	def _create_metacommunity_parameters(self):
		"""
		Creates the METACOMMUNITY_PARAMETERS table in the database.

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a METACOMMUNITY_PARAMETERS table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS METACOMMUNITY_PARAMETERS (reference INT PRIMARY KEY NOT NULL," \
					 " speciation_rate DOUBLE NOT NULL, metacommunity_size DOUBLE NOT NULL);"
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not create COMMUNITY_PARAMETERS table in database {}: {}".format(self.file, soe))

	def _create_species_list(self):
		"""
		Creates the SPECIES_LIST table in the database.

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a SPECIES_LIST table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS SPECIES_LIST (ID int PRIMARY KEY NOT NULL, unique_spec INT NOT NULL, " \
					 "xval INT NOT NULL, yval INT NOT NULL, xwrap INT NOT NULL, ywrap INT NOT NULL, tip INT NOT NULL, " \
					 "speciated INT NOT NULL, " \
					 "parent INT NOT NULL, existence INT NOT NULL, randnum DOUBLE NOT NULL, gen_alive INT NOT NULL, " \
					 "gen_added DOUBLE NOT NULL, guild INT NOT NULL);"
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not create SPECIES_LIST table in database {}: {}".format(self.file, soe))

	def _create_species_locations(self):
		"""
		Creates the SPECIES_LOCATIONS table in the database.

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a SPECIES_LOCATIONS table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS SPECIES_LOCATIONS (ID int PRIMARY KEY NOT NULL, species_id INT " \
					 "NOT NULL, x INT NOT NULL, y INT NOT NULL, community_reference INT NOT NULL, guild INT NOT NULL);"
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not create SPECIES_LOCATIONS table in database {}: {}".format(self.file, soe))

	def _create_species_abundances(self, guilds=False):
		"""
		Creates the SPECIES_ABUNDANCES table in the database.

		:param guilds: if true, creates a separate table for individual guilds, rather than for the whole community

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a SPECIES_ABUNDANCES table as the database has not been created.")
		if guilds:
			additional = "_GUILDS"
		else:
			additional = ""
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS SPECIES_ABUNDANCES{} (ID int PRIMARY KEY NOT NULL, " \
					 "species_id INT NOT NULL, no_individuals INT NOT NULL, community_reference INT NOT NULL, " \
					 "guild INT NOT NULL);".format(additional)
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not create SPECIES_ABUNDANCES table in database {}: {}".format(self.file, soe))

	def _create_fragment_abundances(self, guilds=False):
		"""
		Creates the FRAGMENT_ABUNDANCES table in the database.

		:param guilds: if true, creates a separate table for individual guilds, rather than for the whole community
		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a FRAGMENT_ABUNDANCES table as the database has not been created.")
		self.cursor = self.database.cursor()
		if guilds:
			additional = "_GUILDS"
		else:
			additional = ""
		create_sql = "CREATE TABLE IF NOT EXISTS FRAGMENT_ABUNDANCES{} (ID int PRIMARY KEY NOT NULL," \
					 " fragment TEXT NOT NULL, area DOUBLE NOT NULL, size INT NOT NULL,  species_id INT NOT NULL, " \
					 "no_individuals INT NOT NULL, community_reference int NOT NULL," \
					 " guild INT NOT NULL);".format(additional)
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not create SPECIES_ABUNDANCES table in database {}: {}".format(self.file, soe))

	def _create_species_octaves(self, guilds=False):
		"""
		Creates the SPECIES_OCTAVES table in the database.

		:param guilds: if true, creates a separate table for individual guilds, rather than for the whole community

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a SPECIES_OCTAVES table as the database has not been created.")
		if guilds:
			additional = "_GUILDS"
		else:
			additional = ""
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS SPECIES_OCTAVES{} (ID int PRIMARY KEY NOT NULL," \
					 " fragment TEXT NOT NULL, area DOUBLE NOT NULL, size INT NOT NULL,  species_id INT NOT NULL, " \
					 "no_individuals INT NOT NULL, community_reference int NOT NULL," \
					 " guild INT NOT NULL);".format(additional)
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not create SPECIES_ABUNDANCES table in database {}: {}".format(self.file, soe))

	def _create_fragment_octaves(self, guilds=False):
		"""
		Creates the FRAGMENT_OCTAVES table in the database.

		:param guilds: if true, creates a separate table for individual guilds, rather than for the whole community

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a FRAGMENT_OCTAVES table as the database has not been created.")
		if guilds:
			additional1 = "_GUILDS"
			additional2 = ", guild INT NOT NULL"
		else:
			additional1 = ""
			additional2 = ""
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS FRAGMENT_OCTAVES{} (ref INT PRIMARY KEY NOT NULL, fragment TEXT NOT NULL, " \
					 "octave INT NOT NULL, richness INT NOT NULL{});".format(additional1, additional2)
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not create SPECIES_ABUNDANCES table in database {}: {}".format(self.file, soe))

	def _create_species_richness(self, guilds=False):
		"""
		Creates the SPECIES_RICHNESS table in the database.

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a SPECIES_RICHNESS table as the database has not been created.")
		if guilds:
			additional1 = "_GUILDS"
			additional2 = ", guild INT NOT NULL"
		else:
			additional1 = ""
			additional2 = ""
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE IF NOT EXISTS SPECIES_RICHNESS{} (ref INT PRIMARY KEY NOT NULL," \
					 " community_reference INT NOT NULL, richness INT NOT NULL{})".format(additional1, additional2)

		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not create SPECIES_RICHNESS table in database {}: {}".format(self.file, soe))

	def _create_fragment_richness(self, guilds=False):
		"""
		Creates the FRAGMENT_RICHNESS table in the database.

		:param guilds: if true, creates a separate table for individual guilds, rather than for the whole community

		:raises IOError: if the database has not been successfully created
		"""
		if self.database is None:
			raise IOError("Cannot create a FRAGMENT_RICHNESS table as the database has not been created.")
		if guilds:
			additional1 = "_GUILDS"
			additional2 = ", guild INT NOT NULL"
		else:
			additional1 = ""
			additional2 = ""
		self.cursor = self.database.cursor()
		create_sql = "CREATE TABLE FRAGMENT_RICHNESS{} (ref INT PRIMARY KEY NOT NULL, fragment TEXT NOT NULL," \
					 " community_reference INT NOT NULL,  richness INT NOT NULL{})".format(additional1, additional2)
		try:
			self.cursor.execute(create_sql)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not create FRAGMENT_RICHNESS table in database {}: {}".format(self.file, soe))


	def _read_simulation_parameters(self, input_simulation):
		"""
		Reads the simulation parameters from the input file and returns a list containing these parameters

		:param input_simulation: the simulation to read the simulation parameters from

		:return: list containing the simulation parameters from the input file
		"""
		output = fetch_table_from_sql(input_simulation, "SIMULATION_PARAMETERS")
		if len(output[0]) != 44:
			raise IOError("SIMULATION_PARAMETERS table does not contain 44 columns, contains {}."
						  " Check database.".format(len(output[0])))
		return output[0]

	def _add_simulation_parameters(self, parameters_list, guild, filename):
		"""
		Adds the provided parameters to the output statement for SIMULATION_PARAMETERS with the provided guild reference.
		:param list parameters_list: a list containing the SIMULATION_PARAMETERS outputs.
		:param int guild: the guild reference number for
		:param str filename: the simulation file associated with these parameters
		:return: None
		:rtype: None
		"""
		parameters_list.extend([guild, filename])
		self.parameters_list.append(tuple(parameters_list))

	def _write_simulation_parameters(self):
		"""
		Writes out the parameters_list object to the output database.

		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to SPECIES_PARAMETERS table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO SIMULATION_PARAMETERS VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?," \
					 "?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);"
		try:
			self.cursor.executemany(create_sql, self.parameters_list)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not create SIMULATION_PARAMETERS table in database {}: {}".format(self.file, soe))

	def _read_species_list(self, input_simulation):
		"""
		Reads the SPECIES_LIST table from the input simulation
		:param input_simulation: the completed simulation to read the SPECIES_LIST object from.
		:return:
		"""
		return fetch_table_from_sql(input_simulation, "SPECIES_LIST")

	def _add_species_list(self, species_list, guild):
		"""
		Adds to the species_list object with the provided list of species and guild reference.
		:param species_list: the list of lists of species information to append
		:param guild: the guild to append to each species row.
		"""
		for i, row in enumerate(species_list):
			row[0] = cantor_pairing(row[0], guild)
			if i != 0 and row[8] != 0:
					row[8] = row[8] + self.species_list_parent_counter
			row.append(guild)
		self.species_list_parent_counter += len(species_list)
		self.species_list.extend([tuple(x) for x in species_list])

	def _write_species_list(self):
		"""
		Writes out the species_list object to the output database.

		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to SPECIES_LIST table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO SPECIES_LIST VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?);"
		try:
			self.cursor.executemany(create_sql, self.species_list)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not insert into SPECIES_LIST table in database {}: {}".format(self.file, soe))

	def _read_species_locations(self, input_simulation):
		"""
		Reads the SPECIES_LOCATIONS table from the input simulation
		:param input_simulation: the completed simulation to read the SPECIES_LIST object from.
		:return:
		"""
		return fetch_table_from_sql(input_simulation, "SPECIES_LOCATIONS")

	def _add_species_locations(self, species_locations, guild):
		"""
		Adds to the species_locations object with the provided list of species and guild reference.
		:param species_list: the list of lists of species information to append
		:param guild: the guild to append to each species row.
		"""
		tmp_max = self.max_locations_id
		for row in species_locations:
			row[0] = cantor_pairing(row[0], guild)
			row.append(guild)
			self.max_locations_id = max(row[1], self.max_locations_id)
			row[1] += tmp_max
		self.species_locations.extend([tuple(x) for x in species_locations])

	def _write_species_locations(self):
		"""
		Writes out the species_locations object to the output database.

		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to SPECIES_LIST table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO SPECIES_LOCATIONS VALUES (?,?,?,?,?,?);"
		try:
			self.cursor.executemany(create_sql, self.species_locations)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not insert into SPECIES_LIST table in database {}: {}".format(self.file, soe))

	def _read_species_abundances(self, input_simulation):
		"""
		Reads the SPECIES_ABUNDANCES table from the input simulation.

		:param input_simulation: the completed simulation to read the SPECIES_ABUNDANCES object from.
		:return:
		"""
		return fetch_table_from_sql(input_simulation, "SPECIES_ABUNDANCES")


	def _add_species_abundances(self, species_abundances, guild):
		"""
		Adds to the species_abundances object with the provided list of species and guild reference.

		:param species_abundances: the list of lists of species information to append
		:param guild: the guild to append to each species row.
		"""
		tmp_max = self.max_species_id
		for row in species_abundances:
			row[0] = cantor_pairing(row[0], guild)
			row.append(guild)
			self.max_species_id = max(row[1], self.max_species_id)
			row[1] += tmp_max
		self.species_abundances.extend([tuple(x) for x in species_abundances])

	def _write_species_abundances(self):
		"""
		Writes out the species_list object to the output database.

		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to SPECIES_ABUNDANCES table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO SPECIES_ABUNDANCES VALUES (?,?,?,?,?);"
		try:
			self.cursor.executemany(create_sql, self.species_abundances)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not insert into SPECIES_ABUNDANCES table in database {}: {}".format(self.file, soe))

	def _read_fragment_abundances(self, input_simulation):
		"""
		Reads the FRAGMENT_ABUNDANCES table from the input simulation
		:param input_simulation: the completed simulation to read the FRAGMENT_ABUNDANCES object from.
		:return:
		"""
		return fetch_table_from_sql(input_simulation, "FRAGMENT_ABUNDANCES")

	def _add_fragment_abundances(self, fragment_abundances, guild):
		"""
		Adds to the fragment_abundances object with the provided list of species and guild reference.
		:param fragment_abundances: the list of lists of species information to append
		:param guild: the guild to append to each species row.
		"""
		for row in fragment_abundances:
			row[0] = cantor_pairing(row[0], guild)
			row.append(guild)
		self.fragment_abundances.extend([tuple(x) for x in fragment_abundances])

	def _write_fragment_abundances(self):
		"""
		Writes out the fragment_abundances object to the output database.

		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to FRAGMENT_ABUNDANCES table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO FRAGMENT_ABUNDANCES VALUES (?,?,?,?,?);"
		try:
			self.cursor.executemany(create_sql, self.fragment_abundances)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not insert into FRAGMENT_ABUNDANCES table in database {}: {}".format(self.file, soe))

	def _read_species_richness(self, input_simulation):
		"""
		Reads the SPECIES_RICHNESS table from the input simulation

		:param input_simulation: the completed simulation to read the SPECIES_RICHNESS object from.
		:return:
		"""
		return fetch_table_from_sql(input_simulation, "SPECIES_RICHNESS")

	def _add_species_richness(self, species_richness, guild):
		"""
		Adds to the species_richness object with the provided list of species and guild reference.
		:param species_richness: the list of lists of species information to append
		:param guild: the guild to append to each species row.
		"""
		for row in species_richness:
			row[0] = cantor_pairing(row[0], guild)
			row.append(guild)
		self.species_richness.extend([tuple(x) for x in species_richness])

	def _write_species_richness(self, guilds=False):
		"""
		Writes out the species_richness object to the output database.

		:param guilds: if true, creates a separate table for individual guilds, rather than for the whole community

		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to SPECIES_RICHNESS table as the database has not been created.")
		if guilds:
			additional = "_GUILDS VALUES (?,?,"
		else:
			additional = " VALUES (?,"
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO SPECIES_RICHNESS{} ?,?);".format(additional)
		try:
			self.cursor.executemany(create_sql, self.species_richness)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not insert into SPECIES_RICHNESS table in database {}: {}".format(self.file, soe))

	def _read_fragment_richness(self, input_simulation):
		"""
		Reads the FRAGMENT_RICHNESS table from the input simulation

		:param input_simulation: the completed simulation to read the FRAGMENT_RICHNESS object from.
		:return:
		"""
		return fetch_table_from_sql(input_simulation, "FRAGMENT_RICHNESS")

	def _add_fragment_richness(self, fragment_richness, guild):
		"""
		Adds to the species_richness object with the provided list of species and guild reference.
		:param fragment_richness: the list of lists of species information to append
		:param guild: the guild to append to each species row.
		"""
		for row in fragment_richness:
			row[0] = cantor_pairing(row[0], guild)
			row.append(guild)
		self.fragment_richness.extend([tuple(x) for x in fragment_richness])

	def _write_fragment_richness(self, guilds=False):
		"""
		Writes out the fragment_richness object to the output database.

		:param guilds: if true, creates a separate table for individual guilds, rather than for the whole community


		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to FRAGMENT_RICHNESS table as the database has not been created.")
		if guilds:
			additional = "_GUILDS VALUES (?,"
		else:
			additional = " VALUES ("
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO FRAGMENT_RICHNESS{} ?,?,?,?);".format(additional)
		try:
			self.cursor.executemany(create_sql, self.fragment_richness)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not insert into FRAGMENT_RICHNESS table in database {}: {}".format(self.file, soe))

	def _read_fragment_octaves(self, input_simulation):
		"""
		Reads the FRAGMENT_ABUNDANCES table from the input simulation

		:param input_simulation: the completed simulation to read the FRAGMENT_ABUNDANCES object from.
		:return:
		"""
		try:
			db = sqlite3.connect(input_simulation)
			c = db.cursor()
			c.execute("SELECT ref, fragment, octave, richness FROM FRAGMENT_OCTAVES")
			out = [list(x) for x in c.fetchall()]
			db.close()
			db = None
			return out
		except sqlite3.OperationalError as soe:
			raise IOError("Cannot fetch FRAGMENT_OCTAVES from database {}: {}".format(input_simulation, soe))


	def _add_fragment_octaves(self, fragment_octaves, guild):
		"""
		Adds to the fragment_octaves object with the provided list of species and guild reference.
		:param fragment_octaves: the list of lists of species information to append
		:param guild: the guild to append to each species row.
		"""
		for row in fragment_octaves:
			row[0] = cantor_pairing(row[0], guild)
			row.append(guild)
		self.fragment_octaves.extend([tuple(x) for x in fragment_octaves])

	def _write_fragment_octaves(self, guilds=False):
		"""
		Writes out the fragment_octaves object to the output database.

		:param guilds: if true, creates a separate table for individual guilds, rather than for the whole community

		:return: None
		:rtype: None
		"""
		if self.database is None:
			raise IOError("Cannot insert to FRAGMENT_ABUNDANCES table as the database has not been created.")
		if guilds:
			additional = "_GUILDS VALUES (?,"
		else:
			additional = " VALUES ("
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO FRAGMENT_OCTAVES{} ?,?,?,?);".format(additional)
		try:
			self.cursor.executemany(create_sql, self.fragment_octaves)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not insert into FRAGMENT_ABUNDANCES table in database {}: {}".format(self.file, soe))

	def _add_community_parameters(self, input_simulation):
		"""
		Adds the community parameters to the internal object, checking to ensure that no repeat references exist with
		different parameters.

		:param input_simulation: completed simulation to obtain community references from

		:raises ValueError: if there is a parameter mismatch between the references
		"""
		community_parameters = fetch_table_from_sql(input_simulation, "COMMUNITY_PARAMETERS")
		if len(self.community_parameters) > 0:
			for reference, speciation_rate, time, fragments, metacommunity_reference in community_parameters:
				subsetted_params = [x for x in self.community_parameters if
									isclose(speciation_rate, x[1]) and isclose(time, x[2], abs_tol=0.00001) and
									metacommunity_reference == x[4]]
				if len(subsetted_params) > 0:
					if subsetted_params[0][0] != reference:
						raise ValueError("Parameter mismatch in references for speciation rate = {} and"
										 " time = {}".format(speciation_rate, time))
				else:
					self.logger.error("Parameter set for reference of {} "
										"does not exist in all simulations.".format(reference))
					self.community_parameters.extend(community_parameters)
			for reference, speciation_rate, time, fragments, metacommunity_reference in self.community_parameters:
				subsetted_params = [x for x in community_parameters if isclose(speciation_rate, x[1]) and
									isclose(time, x[2], abs_tol=0.00001) and metacommunity_reference == x[4]]
				if len(subsetted_params) == 0:
					self.logger.error("Parameter set for reference of {} "
									  "does not exist in all simulations.".format(reference))
					self.community_parameters.extend(community_parameters)
		else:
			self.community_parameters.extend(community_parameters)

	def _write_community_parameters(self):
		"""
		Outputs the community parameters into the database
		"""
		if self.database is None:
			raise IOError("Cannot insert to COMMUNITY_PARAMETERS table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO COMMUNITY_PARAMETERS VALUES (?,?,?,?,?);"
		try:
			self.cursor.executemany(create_sql, self.community_parameters)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not insert into COMMUNITY_PARAMETERS table in database {}: {}".format(self.file, soe))

	def _add_metacommunity_parameters(self, input_simulation):
		"""
		Adds the metacommunity parameters to the internal object, checking to ensure that no repeat references exist
		with different parameters.

		:param input_simulation: completed simulation to obtain metacommunity references from

		:raises ValueError: if there is a parameter mismatch between the references
		"""
		metacommunity_parameters = fetch_table_from_sql(input_simulation, "METACOMMUNITY_PARAMETERS")
		if len(self.community_parameters) > 0:
			for reference, speciation_rate, metacommunity_size in metacommunity_parameters:
				subsetted_params = [x for x in self.metacommunity_parameters if
									isclose(speciation_rate, x[1]) and metacommunity_size == x[2]]
				if len(subsetted_params) > 0:
					if subsetted_params[1] != reference:
						raise ValueError("Parameter mismatch in metacommunity references for speciation rate = {} and"
										 " size = {}".format(speciation_rate, metacommunity_size))
				else:
					self.logger.error("Parameter set for metacommunity reference of {} "
									  "does not exist in all simulations.".format(reference))
					self.metacommunity_parameters.extend(metacommunity_parameters)
			for reference, speciation_rate, metacommunity_size in self.metacommunity_parameters:
				subsetted_params = [x for x in metacommunity_parameters if isclose(speciation_rate, x[1]) and
									metacommunity_size == x[2]]
				if len(subsetted_params) == 0:
					self.logger.error("Parameter set for metacommunity reference of {} "
									  "does not exist in all simulations.".format(reference))
					self.metacommunity_parameters.extend(metacommunity_parameters)
		else:
			self.metacommunity_parameters.extend(metacommunity_parameters)

	def _write_metacommunity_parameters(self):
		if self.database is None:
			raise IOError("Cannot insert to METACOMMUNITY_PARAMETERS table as the database has not been created.")
		self.cursor = self.database.cursor()
		create_sql = "INSERT INTO METACOMMUNITY_PARAMETERS VALUES (?,?,?);"
		try:
			self.cursor.executemany(create_sql, self.community_parameters)
			self.database.commit()
		except sqlite3.OperationalError as soe:
			raise IOError("Could not insert into METACOMMUNITY_PARAMETERS table in database {}: {}".format(self.file, soe))

	def add_simulation(self, input_simulation):
		"""
		Adds a simulation to the list of merged simulations.

		This also calls the relevant merges for the tables that exist in the provided database.

		:param input_simulation: either the path to the input simulation, a Coalescence class object, or a CoalescenceTree object
			which contains the completed simulation.
		:return: None
		:rtype: None
		"""
		if isinstance(input_simulation, CoalescenceTree):
			t = input_simulation
		else:
			t = CoalescenceTree(input_simulation)
		simulation_path = t.file
		# Ensure that any open sql connections are closed
		del t
		self.guild += 1
		# First add the parameters (which might throw some simple errors)
		self._add_simulation_parameters(self._read_simulation_parameters(simulation_path), self.guild, simulation_path)
		if check_sql_table_exist(simulation_path, "COMMUNITY_PARAMETERS"):
			self._create_community_parameters()
			self._add_community_parameters(simulation_path)
		if check_sql_table_exist(simulation_path, "METACOMMUNITY_PARAMETERS"):
			self._create_metacommunity_parameters()
			self._add_metacommunity_parameters(simulation_path)
		# Now the longer routines
		self._add_species_list(self._read_species_list(simulation_path), self.guild)
		if check_sql_table_exist(simulation_path, "SPECIES_ABUNDANCES"):
			self._create_species_abundances()
			self._add_species_abundances(self._read_species_abundances(simulation_path), self.guild)
		if check_sql_table_exist(simulation_path, "FRAGMENT_ABUNDANCES"):
			self._create_fragment_abundances()
			self._add_fragment_abundances(self._read_fragment_abundances(simulation_path), self.guild)
		if check_sql_table_exist(simulation_path, "SPECIES_RICHNESS"):
			self._create_species_richness(guilds=True)
			self._add_species_richness(self._read_species_richness(simulation_path), self.guild)
		if check_sql_table_exist(simulation_path, "FRAGMENT_RICHNESS"):
			self._create_fragment_richness(guilds=True)
			self._add_fragment_richness(self._read_fragment_richness(simulation_path), self.guild)
		if check_sql_table_exist(simulation_path, "FRAGMENT_OCTAVES"):
			self._create_fragment_octaves(guilds=True)
			self._add_fragment_octaves(self._read_fragment_octaves(simulation_path), self.guild)

	def add_simulations(self, simulation_list):
		"""
		A convenience function that adds each simulation from the list of simulations provided and then writes to the
		database.

		:param simulation_list: list of paths to completed simulations
		"""
		for each in simulation_list:
			self.add_simulation(each)
		self.write()

	def write(self):
		"""
		Writes out all stored simulation parameters to the output database and wipes the in-memory objects.

		This should be called after all simulation have been added, or when RAM usage gets too large for large
		simulations

		"""
		self._write_simulation_parameters()
		if len(self.community_parameters) > 0:
			self._write_community_parameters()
		if len(self.metacommunity_parameters) > 0:
			self._write_metacommunity_parameters()
		self._write_species_list()
		if len(self.species_abundances) > 0:
			self._write_species_abundances()
		if len(self.fragment_abundances) > 0:
			self._write_fragment_abundances()
		if len(self.species_richness) > 0:
			self._write_species_richness(guilds=True)
			self._create_combined_species_richness()
		if len(self.fragment_richness) > 0:
			self._write_fragment_richness(guilds=True)
		if len(self.fragment_octaves) > 0:
			self._write_fragment_octaves(guilds=True)

	def _create_combined_species_richness(self):
		"""
		Combines the species richness guilds so that richness values are for all guilds.
		"""
		if len(self.species_richness) == 0:
			raise ValueError("No species richnesses are detected. Add simulations before combining richness.")
		self._create_species_richness(guilds=False)
		out_richness = []
		references = set([x[1] for x in self.species_richness])
		i = 0
		for reference in references:
			total = sum([x[2] for x in self.species_richness if x[1] == reference])
			out_richness.append([i, reference, total])
			i += 1
		self.species_richness = out_richness
		self._write_species_richness(guilds=False)

	def _create_combined_fragment_richness(self):
		"""
		Combines the fragment richness guilds so that richness values are for all guilds in each fragment.
		"""
		if len(self.fragment_richness) == 0:
			raise ValueError("No fragment richnesses are detected. Add simulations before combining richness.")
		self._create_fragment_richness(guilds=False)
		out_richness = []
		fragments = set([x[1] for x in self.fragment_richness])
		references = set([x[2] for x in self.fragment_richness])
		i = 0
		for reference in references:
			for fragment in fragments:
				total = sum([x[3] for x in self.fragment_richness if x[1] == fragment and x[2] == reference])
				out_richness.append([i, fragment, reference, total])
				i += 1
		self.fragment_richness = out_richness
		self._write_fragment_richness(guild=False)

	def _create_combined_fragment_octaves(self):
		"""
		Combines the fragment octaves guilds so that octaves values are for all guilds in each fragment.
		"""
		if len(self.fragment_octaves) == 0:
			raise ValueError("No fragment octaves are detected. Add simulations before creating combined octaves")
		self._create_fragment_octaves(guilds=False)
		octaves = []
		fragments = set([x[1] for x in self.fragment_octaves])
		references = set([x[2] for x in self.fragment_octaves])
		i = 0
		for reference in references:
			for fragment in fragments:
				total = sum([x[3] for x in self.fragment_octaves if x[1] == fragment and x[2] == reference])
				octaves.append([i, fragment, reference, total])
				i += 1
		self.fragment_octaves = octaves
		self._write_fragment_octaves(guild=False)




"""
Compile **necsim** with a number of intel compiler optimisations for running on high-performance computing systems.
"""
from __future__ import absolute_import
import logging

try:
	from .installer import Installer
except (ImportError, SystemError, ValueError):
	from installer import Installer


def build_hpc():
	"""
	Compiles NECSim with the ``--with-hpc`` and ``--with-verbose`` flags, which adds extra support for intel compilers
	and provides a selection of optimisation flags for high-performance systems.
	:return:
	"""
	from distutils.dist import Distribution
	dist = Distribution()
	installer = Installer(dist)

	installer.configure_and_compile(argv=["--with-hpc", "--with-verbose"], logging_level=logging.INFO)


if __name__ == "__main__":
	build_hpc()



"""
Basic system-level operations required for package functionality, including subprocess calls, logging methods and file
management.

The functions are contained here as they are required by many different modules. Note that logging will not raise an
exception if there has been no call to set_logging_method()

"""

import logging
import os
import subprocess
import sys

# Logging will not raise an exception if there has been no logging file set.
logging.raiseExceptions = True
logging_set = False
mod_directory = os.path.dirname(os.path.abspath(__file__))

try:
	from math import isclose
except ImportError:
	# Python 2 compatibility
	def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
		"""
		Checks if the two floats are close.

		:param float a: value 1
		:param float b: value 2
		:param float rel_tol: percentage relative to larger value
		:param float abs_tol: absolute value for similarity

		:return: true for significantly different a and b, false otherwise
		"""
		return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def check_parent(file_path):
	"""
	Checks if the parent file exists, and creates it if it doesn't.

	.. note:: if file_path is a directory, it will be created

	:param file_path: the file or directory to check if the parent exists

	:rtype: None
	"""
	if file_path:
		if not os.path.exists(file_path):
			if os.path.isdir(file_path):
				os.makedirs(file_path)
			else:
				par_dir = os.path.dirname(file_path)
				if not os.path.exists(par_dir) and par_dir != "":
					os.makedirs(os.path.dirname(file_path))
	else:
		raise ValueError("File path is None")


def check_file_exists(file_name):
	"""
	Checks that the specified filename exists, if it is not "null" or "none".

	:param file_name: file path to check for

	:return: None

	:raises: IOError if no file exists
	"""
	if file_name not in {"null", "none", None} and not os.path.exists(file_name):
		err = "Map file " + file_name + " does not exist. Check that this is an absolute path or disable map " \
										"file check"
		raise IOError(err)


def execute(cmd, silent=False, **kwargs):
	"""
	Calls the command using subprocess and yields the running output for printing to terminal. Any errors produced by
	subprocess call will be redirected to logging.warning() after the subprocess call is complete.

	:param cmd: the command to execute using subprocess.Popen()
	:param silent: if true, does not log any warnings

	:return a line from the execution output
	"""
	popen = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
							 universal_newlines=True, **kwargs)
	while True:
		output = popen.stdout.readline()
		if output == '' and popen.poll() is not None:
			break
		if output:
			yield output
	error_logs = []
	for stderr_line in iter(popen.stderr.readline, ""):
		error_logs.append(stderr_line)
	popen.stdout.close()
	popen.stderr.close()
	popen.wait()
	if popen.returncode != 0:
		for log in error_logs:
			logging.critical(log)
		raise RuntimeError(error_logs[-1])
	else:
		if not silent:
			for log in error_logs:
				logging.warning(log)


def execute_log_info(cmd, **kwargs):
	"""
	Calls execute() with the supplied command and keyword arguments, and redirects stdout to the logging object.

	:param cmd: the command to execute using subprocess.Popen()
	:param kwargs: keyword arguments to be passed to subprocess.Popen()

	:return: None
	:rtype: None
	"""
	for line in execute(cmd, **kwargs):
		logging.info(line.strip("\n"))


def execute_silent(cmd, **kwargs):
	"""
	Calls execute() silently with the supplied command and keyword arguments.

	.. note:: If this function fails, no error will be thrown due to its silent nature, unless a full failure occurs.

	:param cmd: the command to execute using subprocess.Popen()
	:param kwargs: keyword arguments to be passed to subprocess.Popen()

	:return: None
	:rtype: None
	"""
	for _ in execute(cmd, silent=True, **kwargs):
		continue


def set_logging_method(logging_level=logging.INFO, output=None, **kwargs):
	"""
	Initiates the logging process.

	:param logging_level: the detail in logging output: can be one
	 				     of logging.INFO (default), logging.WARNING, logging.DEBUG, logging.ERROR or logging.CRITICAL
	:param output: the output logfile (or None to redirect to terminal via stdout)
	:param kwargs: additional arguments to pass to the logging.basicConfig() call

	:return: None
	"""
	if not logging_set:
		if output is None:
			logging.basicConfig(stream=sys.stdout, format='%(levelname)s:%(message)s', level=logging_level, **kwargs)
		else:
			logging.basicConfig(filename=output, format='%(levelname)s:%(message)s', level=logging_level, **kwargs)
		logging.captureWarnings(True)


def write_to_log(i, message, logger):
	"""
	Writes the message to the provided logger, at the provided level.

	This is used by NECSim to access to logging module more easily.

	:param i: the level to log at (10: debug, 20: info, 30: warning, 40: error, 50: critical)
	:param message: the message to write to the logger.
	:param logger:

	:type i: int
	:type message: str
	:type logger: logging.Logger

	:rtype: None
	"""
	logger.log(level=i, msg=message)


def create_logger(logger, file=None, logging_level=logging.WARNING, **kwargs):
	"""
	Creates a logger object to be assigned to NECSim sims and dispersal tests.

	:param logger: the logger to alter
	:param file: the file to write out to, defaults to None, writing to terminal
	:param logging_level: the logging level to write out at (defaults to INFO)
	:param kwargs: optionally provide additional arguments for logging to

	:return:
	"""
	formatter = logging.Formatter('%(message)s')
	if file is None:
		sh = logging.StreamHandler(kwargs.get("stream", None))
	else:
		sh = logging.FileHandler(file)
	sh.setLevel(logging_level)
	sh.setFormatter(formatter)
	if sys.version[0] == "3":
		sh.terminator = ""
	logger.addHandler(sh)
	return logger


def elegant_pairing(x1, x2):
	"""
	A more elegant version of cantor pairing, which allows for storing of a greater number of digits without
	experiencing integer overflow issues.

	Cantor pairing assigns consecutive numbers to points along diagonals of a plane

	:param x1: the first number
	:param x2: the second number

	:return: a unique reference combining the two integers.
	"""
	if x1 > x2:
		return x1 ** 2 + x1 + x2
	return x2 ** 2 + x1


def cantor_pairing(x1, x2):
	"""
	Creates a unique integer from the two provided positive integers.

	Maps ZxZ -> N, so only relevant for positive numbers.
	For any A and B, generates C such that no D and E produce C unless D=A and B=E.

	Assigns consecutive numbers to points along diagonals of a plane

	.. note:
		For most use cases which are not performance-critical, :func:`~elegant_pairing` provides a more reliable outcome
		by reducing the size of the integers and therefore reducing the change of an integer overflow.

	:param x1: the first number
	:param x2: the second number

	:return: a unique reference combining the two integers
	"""
	return ((x1 + x2) * (x1 + x2 + 1) / 2) + x2



"""
Run spatially-explicit neutral simulations on provided landscapes with support for a wide range of scenarios and
parameters. Detailed :ref:`here <performing_simulations>`.

The main class is :class:`.Simulation`, which contains routines for setting up and running simulations, plus basic tree
generation after simulations have been completed.

:input:
	- Simulation parameters (such as dispersal kernel, speciation rate)
	- Map files representing density over space
	- [optional] map files representing relative reproductive ability
	- [optional] map files representing dispersal potential
	- [optional] historical density map files

:output:
	- Database containing generated coalescence tree, simulation parameters and basic biodiversity metrics.
	- If the simulation does not complete, will instead generate a set of Dump_*.csv files for resuming simulations
"""
from __future__ import absolute_import
from __future__ import print_function

import copy
import logging
import os
import sys
import types

import numpy as np

global sqlite_import
if sys.version_info[0] == 3:
	import configparser as ConfigParser
	from io import StringIO
else:
	import ConfigParser
	from cStringIO import StringIO

	ConfigParser.ConfigParser.read_file = ConfigParser.ConfigParser.read

try:
	NumberTypes = (types.IntType, types.LongType, types.FloatType, types.ComplexType)
except AttributeError:
	# No support for complex numbers compile
	NumberTypes = (int, float)

try:
	from osgeo import gdal
except ImportError as ie:
	try:
		import gdal
	except ImportError:
		gdal = None
		logging.warning("Problem importing  (gdal) modules. " + str(ie))

try:
	import sqlite3

	sqlite_import = True
	from .coalescence_tree import CoalescenceTree
except ImportError as ie:
	sqlite_import = False
	CoalescenceTree = None
	sqlite3 = None
	logging.warning("Problem importing sqlite module " + str(ie))

from pycoalescence.necsim import libnecsim
from pycoalescence.future_except import FileNotFoundError, FileExistsError
from pycoalescence.landscape import Landscape
from pycoalescence.map import Map
from pycoalescence.system_operations import write_to_log, check_parent


class Simulation(Landscape):
	"""
	A class containing routines to set up and run simulations, including detecting map dimensions from tif files.
	"""

	def __init__(self, logging_level=logging.WARNING,
				 log_output=None, **kwargs):
		"""
		Sets up the simulation object, setting the logging object and assigning default variables.

		:param int logging_level: the minimal level of logging to display
		:param string log_output: a file to output log data to. Use None for writing to console.
		:param kwargs: additional keyword arguments to supply to logging.basicConfig()
		"""
		# All the parameters used for the simulation
		Landscape.__init__(self)
		# Contains the C object for performing simulations
		# This will be set depending on the simulation type
		self.c_simulation = None
		self._fine_map_sum_res = None
		self.count_total = None
		# Create the logging object and pass on the required arguments for its creation
		passedKwards = {k: v for k, v in kwargs.items() if "stream" not in k}
		self.logger = logging.Logger("pycoalescence.simulation", **passedKwards)
		self._create_logger(file=log_output, logging_level=logging_level, **kwargs)
		self.output_directory = ""
		self.pause_directory = ""
		self.config = ConfigParser.ConfigParser()
		self.config_string = StringIO()
		self.full_config_file = None
		self.is_setup = False
		self.is_setup_param = False
		self.is_setup_complete = False
		self.seed = 0
		self.min_speciation_rate = 0
		self.sigma = 2
		self.deme = 1
		self.sample_size = 1
		self.max_time = 100
		self.dispersal_relative_cost = 1  # the relative cost of moving through non-matrix
		self.job_type = 0
		self.min_num_species = 1
		self.tau = 0
		self.times = None
		# the optional parameters specifying the location (i.e. where in the world the simulation is of
		# and the compute node (such as NUS_HPC or LOCAL_MACBOOK)
		self.sim_location = None
		self.sim_compute_node = None
		self.speciation_rates = []
		# parameters for the application of speciation rates
		self.is_setup_speciation = False
		self.speciation_file = ""
		self.record_spatial = False
		self.record_fragments = False
		# The output database location
		self.output_database = None
		self.dispersal_method = None
		self.m_prob = 0
		self.cutoff = 0
		self.restrict_self = False
		# The protracted variables
		self.protracted = False
		self.min_speciation_gen = 0.0
		self.max_speciation_gen = 0.0
		# This variable stores whether a new config file has been opened or not
		self.config_open = False
		# Grid dimensions sizing
		self.grid = Map()
		# Dispersal map data
		self.dispersal_map = Map()
		self.reproduction_map = Map()
		self.fine_map_array = None
		self.sample_map_array = None
		self.fine_map_average = None
		self.is_spatial = True;
		self.uses_spatial_sampling = False

	def __del__(self):
		"""Safely destroys the logger and the c++ objects."""
		handlers = copy.copy(self.logger.handlers)
		for handler in handlers:
			handler.close()
			self.logger.removeHandler(handler)
		self.logger = None
		self.c_simulation = None

	def setup_necsim(self):
		"""
		Sets the logging function and the logger object for the necsim object. Enforcing this function is always called
		ensures no seg faults occur.
		"""
		self.necsim.set_logger(self.logger)
		self.necsim.set_log_function(write_to_log)

	def add_sample_time(self, time):
		"""
		Adds an extra sample time to the list of times.

		This allows for multiple temporal sample points from within the same simulation.

		:param time: the sample time to add
		"""
		if isinstance(time, list):
			[self.add_sample_time(x) for x in time]
		else:
			if not self.times or len(self.times) == 0:
				self.times = [0.0]
			self.times.append(time)

	def set_config_file(self, output_file=None):
		"""
		Sets the config file to the output, over-writing any existing config file that has been stored.

		:param output_file: path to config file to output to
		"""
		if self.full_config_file and output_file != self.full_config_file and output_file:
			self.logger.warning("Config file has already been set to {}, changing to {}.".format(self.full_config_file,
																								 output_file))
		if output_file is not None:
			self.full_config_file = output_file

	def load_config(self, config_file):
		"""
		Loads the config file by reading the lines in order.

		:param str config_file: the config file to read in.
		"""
		# New method using ConfigParser
		with open(config_file, "wa") as f:
			self.config.read_file(f)
		self.seed = self.config.getint("main", "seed")
		self.job_type = self.config.getint("main", "job_type")
		self.output_directory = self.config.get("main", "output_directory")
		self.min_speciation_rate = self.config.getfloat("main", "min_spec_rate")
		self.sigma = self.config.getfloat("main", "sigma")
		self.tau = self.config.getfloat("main", "tau")
		self.deme = self.config.getint("main", "deme")
		self.sample_size = self.config.getfloat("main", "sample_size")
		self.max_time = self.config.getint("main", "max_time")
		self.dispersal_relative_cost = self.config.getfloat("main", "lambda")
		self.min_num_species = self.config.getint("main", "min_species")
		self.full_config_file = config_file
		if self.config.has_section("spec_rate"):
			opts = self.config.options("spec_rate")
			for each in opts:
				self.speciation_rates.append(self.config.getfloat("spec_rate", each))

	def write_config(self, config_file):
		"""
		Writes the config to the config file provided, overwriting any existing config files.

		:param config_file: the config file to write out to

		:rtype: None
		"""
		if config_file is not None:
			check_parent(config_file)
			if os.path.exists(config_file) and not self.config_open:
				os.remove(config_file)
		with open(config_file, "w") as f:
			self.config.write(f)
		self.config_open = True
		self.full_config_file = config_file

	def create_temporal_sampling_config(self):
		"""
		Creates the time-sampling config file.

		Function is called automatically when creating a config file, and should not be manually called.
		"""
		if self.times and len(self.times) != 0:
			self.config.add_section("times")
			for i, j in enumerate(sorted(set(self.times))):
				self.config.set("times", "time" + str(i), str(j))
			if self.full_config_file is not None:
				self.write_config(self.full_config_file)

	def create_map_config(self, output_file=None):
		"""
		Generates the map config file from reading the spatial structure of each of the provided files.

		:param str output_file: the file to output configuration data to (the map config file)
		"""
		self.set_config_file(output_file)
		if self.is_setup_map:
			if self.is_spatial:
				self.config.add_section("sample_grid")
				self.config.set("sample_grid", "path", self.sample_map.file_name)
				x, y = map(str, self.sample_map.get_x_y())
				self.config.set("sample_grid", "x", str(x))
				self.config.set("sample_grid", "y", str(y))
				if self.sample_map.file_name is None:
					self.config.set("sample_grid", "mask", "null")
				else:
					self.config.set("sample_grid", "mask", self.sample_map.file_name)
					self.config.set("sample_grid", "uses_spatial_sampling", str(int(self.uses_spatial_sampling)))
				self.config.add_section("fine_map")
				self.config.set("fine_map", "path", self.fine_map.file_name)
				x, y = map(str, self.fine_map.get_x_y())
				self.config.set("fine_map", "x", str(x))
				self.config.set("fine_map", "y", str(y))
				self.config.set("fine_map", "x_off", str(self.fine_map.x_offset))
				self.config.set("fine_map", "y_off", str(self.fine_map.y_offset))
				self.config.add_section("coarse_map")
				if self.coarse_map.file_name is not None:
					self.config.set("coarse_map", "path", self.coarse_map.file_name)
					x, y = map(str, self.coarse_map.get_x_y())
					self.config.set("coarse_map", "x", str(x))
					self.config.set("coarse_map", "y", str(y))
					self.config.set("coarse_map", "x_off", str(self.coarse_map.x_offset))
					self.config.set("coarse_map", "y_off", str(self.coarse_map.y_offset))
					self.config.set("coarse_map", "scale", str(self.coarse_scale))
				else:
					self.config.set("coarse_map", "path", "none")
					self.config.set("coarse_map", "x", 0)
					self.config.set("coarse_map", "y", 0)
					self.config.set("coarse_map", "x_off", 0)
					self.config.set("coarse_map", "y_off", 0)
					self.config.set("coarse_map", "scale", 1)
				self.sort_historical_maps()
				if len(self.historical_coarse_list) != 0 and len(self.historical_fine_list) != 0:
					for i, t in enumerate(self.times_list):
						try:
							tmp_fine = "historical_fine{}".format(i)
							self.config.add_section(tmp_fine)
							self.config.set(tmp_fine, "path", self.historical_fine_list[i])
							self.config.set(tmp_fine, "number", str(i))
							self.config.set(tmp_fine, "time", str(t))
							self.config.set(tmp_fine, "rate", str(self.rates_list[i]))
							tmp_coarse = "historical_coarse{}".format(i)
							self.config.add_section(tmp_coarse)
							self.config.set(tmp_coarse, "path", self.historical_coarse_list[i])
							self.config.set(tmp_coarse, "number", str(i))
							self.config.set(tmp_coarse, "time", str(t))
							self.config.set(tmp_coarse, "rate", str(self.rates_list[i]))
						except IndexError as ie:
							self.logger.warning(
								'Discrepancy between historical file list, time list or rate list. Check inputs: {}'.format(
									ie))
							break
				if self.grid.file_name == "set":
					self.config.add_section("grid_map")
					self.config.set("grid_map", "x", str(self.grid.x_size))
					self.config.set("grid_map", "y", str(self.grid.y_size))
					self.config.set("sample_grid", "x_off", str(self.sample_map.x_offset))
					self.config.set("sample_grid", "y_off", str(self.sample_map.y_offset))
				if self.reproduction_map.file_name not in [None, "none", "null"]:
					self.config.add_section("reproduction")
					self.config.set("reproduction", "map", self.reproduction_map.file_name)
				if self.dispersal_method is not None:
					self.config.add_section("dispersal")
					self.config.set("dispersal", "method", self.dispersal_method)
					self.config.set("dispersal", "m_probability", str(self.m_prob))
					self.config.set("dispersal", "cutoff", str(self.cutoff))
				if self.restrict_self:
					if not self.config.has_section("dispersal"):
						self.config.add_section("dispersal")
					self.config.set("dispersal", "restrict_self", str(int(self.restrict_self)))
				if self.landscape_type != "closed":
					if not self.config.has_section("dispersal"):
						self.config.add_section("dispersal")
					self.config.set("dispersal", "landscape_type", self.landscape_type)
				if self.dispersal_map.file_name not in ["none", None]:
					if not self.config.has_section("dispersal"):
						self.config.add_section("dispersal")
					self.config.set("dispersal", "dispersal_file", self.dispersal_map.file_name)
				if self.full_config_file is not None:
					self.write_config(self.full_config_file)
		else:
			raise RuntimeError("Cannot generate map config file without setting up map variables")

	def create_config(self, output_file=None):
		"""
		Generates the configuration. This will be written out either by providing an output file here, or by calling
		write_config();

		:param str output_file: the file to generate the config option. Must be a path to a .txt file.
		"""
		self.set_config_file(output_file)
		if self.full_config_file is not None:
			if not os.path.exists(os.path.dirname(self.full_config_file)) and \
					os.path.dirname(self.full_config_file) != "":
				self.logger.info('Path {} does not exist for writing '
								 'output to, creating.'.format(os.path.dirname(self.full_config_file)))
				os.makedirs(os.path.dirname(self.full_config_file))
			if os.path.exists(self.full_config_file) and not self.config_open:
				os.remove(self.full_config_file)
		if self.is_setup_map and self.is_setup_param:
			# New method using ConfigParser
			self.config.add_section("main")
			self.config.set("main", "seed", str(self.seed))
			self.config.set("main", "job_type", str(self.job_type))
			self.config.set("main", "output_directory", self.output_directory)
			self.config.set("main", "min_spec_rate", str(self.min_speciation_rate))
			if self.is_spatial:
				self.config.set("main", "sigma", str(self.sigma))
				self.config.set("main", "tau", str(self.tau))
				self.config.set("main", "dispersal_relative_cost", str(self.dispersal_relative_cost))
			self.config.set("main", "deme", str(self.deme))
			self.config.set("main", "sample_size", str(self.sample_size))
			self.config.set("main", "max_time", str(self.max_time))
			self.config.set("main", "min_species", str(self.min_num_species))
			if self.protracted:
				self.config.add_section("protracted")
				self.config.set("protracted", "has_protracted", str(int(self.protracted)))
				self.config.set("protracted", "min_speciation_gen", str(float(self.min_speciation_gen)))
				self.config.set("protracted", "max_speciation_gen", str(float(self.max_speciation_gen)))
			if len(self.speciation_rates) != 0:
				self.config.add_section("spec_rates")
				for i, j in enumerate(set([x for x in self.speciation_rates])):
					spec_rate = "spec_{}".format(i)
					self.config.set("spec_rates", spec_rate, str(j))
			if self.full_config_file is not None:
				self.write_config(self.full_config_file)
		else:
			raise RuntimeError("Setup has not been completed, cannot create config file")
		self.create_temporal_sampling_config()

	def run_simple(self, seed, task, output, alpha, sigma, size):
		"""
		Runs a simple coalescence simulation on a square infinite landscape with the provided parameters.
		This requires a separate compilation of the inf_land version of the coalescence simulator.

		Note that this function returns richness=0 for failure to read from the file. It is assumed that there will
		be at least one species in the simulation.

		Note that the maximum time for this function is set as 10 hours (36000 seconds) and will raise an exception if
		the simulation does not complete in this time).

		:raises RuntimeError: if the simulation didn't complete in time.

		:param seed: the simulation seed
		:param task: the task (for file naming)
		:param output: the output directory
		:param alpha: the speciation rate
		:param sigma: the normal distribution sigma value for dispersal
		:param size: the size of the world (so there will be size^2 individuals simulated)

		:return: the species richness in the simulation
		"""
		self.set_simulation_params(seed=seed, job_type=task, output_directory=output, min_speciation_rate=alpha,
								   sigma=sigma, tau=1, deme=1, sample_size=1.0, max_time=36000,
								   dispersal_relative_cost=1,
								   min_num_species=1, habitat_change_rate=0.0, gen_since_historical=1)
		self.set_map_parameters("null", size, size, "null", size, size, 0, 0, "null", size, size, 0, 0, 1, "null",
								"null")
		self.set_speciation_rates([alpha])
		self.run()
		# Now read the species richness from the database
		try:
			richness = self.get_richness()
			if richness == 0:
				raise AssertionError("Richness is 0, error in program. This should never be the case for a complete "
									 "sim.")
			return richness
		except IOError:
			raise RuntimeError("Simulation didn't complete in 10 hours (maximum time for run_simple). Try running a "
							   "custom simulation instead.")

	def get_richness(self, reference=1):
		"""
		Calls coal_analyse.get_richness() with the supplied variables.

		Requires successful import of coal_analyse and sqlite3.

		:param speciation_rate: the speciation rate to extract system richness from.
		:param time: the time to extract system richness from

		:return: the species richness.
		"""
		if not sqlite_import:
			raise ImportError("sqlite3 module required for obtaining richness from database files.")
		else:
			self.calculate_sql_database()
			db = self.output_database
			t = CoalescenceTree()
			t.set_database(db)
			return t.get_richness(reference)

	def get_protracted(self):
		"""
		Gets whether the simulation pointed to by this object is a protracted simulation or not.
		"""
		if self.output_database is None or not os.path.exists(self.output_database):
			return self.protracted
		else:
			t = CoalescenceTree()
			try:
				t.set_database(self.output_database)
			except IOError:
				# Catches errors thrown if the simulation is paused
				pass
			return t.is_protracted()

	def set_map_files(self, sample_file, fine_file=None, coarse_file=None, historical_fine_file=None,
					  historical_coarse_file=None, dispersal_map=None, reproduction_map=None):
		"""
		Sets the map files (or to null, if none specified). It then calls detect_map_dimensions() to correctly read in
		the specified dimensions.

		If sample_file is "null", dimension values will remain at 0.
		If coarse_file is "null", it will default to the size of fine_file with zero offset.
		If the coarse file is "none", it will not be used.
		If the historical fine or coarse files are "none", they will not be used.

		.. note:: the dispersal map should be of dimensions xy by xy where x, y are the fine map dimensions. Dispersal
				  probabilities should sum to 1 across each row, and each row/column index represents dispersal from the
				  row index to the column index according to index = x+(y*xdim), where x,y are the coordinates of the
				  cell and xdim is the x dimension of the fine map. See the
				  :class:`PatchedLandscape class <pycoalescence.patched_landscape.PatchedLandscape>` for routines for
				  generating these landscapes.

		:param str sample_file: the sample map file. Provide "null" if on samplemask is required
		:param str fine_file: the fine map file. Defaults to "null" if none provided
		:param str coarse_file: the coarse map file. Defaults to "none" if none provided
		:param str historical_fine_file: the historical fine map file. Defaults to "none" if none provided
		:param str historical_coarse_file: the historical coarse map file. Defaults to "none" if none provided
		:param str dispersal_map: the dispersal map for reading dispersal values. Default to "none" if none provided
		:param str reproduction_map: a map of relative reproduction probabilities, at the scale of the fine map

		:rtype: None

		:return: None
		"""
		if dispersal_map is None:
			self.dispersal_map.file_name = "none"
		else:
			self.dispersal_map.file_name = dispersal_map
		if reproduction_map is None:
			self.reproduction_map.file_name = "none"
		else:
			self.reproduction_map.file_name = reproduction_map
		Landscape.set_map_files(self, sample_file, fine_file, coarse_file, historical_fine_file, historical_coarse_file)

	def detect_map_dimensions(self):
		"""
		Detects all the map dimensions for the provided files (where possible) and sets the respective values.
		This is intended to be run after set_map_files()

		:raises TypeError: if a dispersal map or reproduction map is specified, we must have a fine map specified, but
						   not a coarse map.

		:raises IOError: if one of the required maps does not exist
		
		:raises ValueError: if the dimensions of the dispersal map do not make sense when used with the fine map
							provided

		:return: None
		"""
		Landscape.detect_map_dimensions(self)
		# Now do detection for dispersal map
		if self.dispersal_map.file_name not in {"none", "null", None}:
			self.dispersal_map.set_dimensions()
			if self.dispersal_map.x_size != self.fine_map.x_size * self.fine_map.y_size or \
					self.dispersal_map.y_size != self.fine_map.x_size * self.fine_map.y_size:
				raise ValueError("Dimensions of dispersal map do not match dimensions of fine map. This is currently"
								 " unsupported.")
		if self.reproduction_map.file_name not in {"none", "null", None}:
			self.reproduction_map.set_dimensions()
			if not self.reproduction_map.has_equal_dimensions(self.fine_map):
				# if the sizes match, then proceed with a warning
				if self.reproduction_map.x_size == self.fine_map.x_size and \
						self.reproduction_map.y_size == self.fine_map.y_size:
					self.logger.warning("Coordinates of reproduction map did not match fine map.")
					self.reproduction_map.x_offset = self.fine_map.x_offset
					self.reproduction_map.y_offset = self.fine_map.y_offset
					self.reproduction_map.x_res = self.fine_map.x_res
					self.reproduction_map.y_res = self.fine_map.y_res
				else:
					raise ValueError("Dimensions of the reproduction map do not match the fine map. This is currently "
									 "unsupported.")
		self.check_maps()

	def set_simulation_params(self, seed, job_type, output_directory, min_speciation_rate, sigma=1.0, tau=1.0, deme=1,
							  sample_size=1.0, max_time=3600, dispersal_method=None, m_prob=0.0, cutoff=0,
							  dispersal_relative_cost=1, min_num_species=1, habitat_change_rate=0.0,
							  gen_since_historical=1, restrict_self=False, landscape_type=False,
							  protracted=False, min_speciation_gen=None, max_speciation_gen=None,
							  spatial=True, uses_spatial_sampling=False, times=None):
		"""
		Set all the simulation parameters apart from the map objects.

		:param int seed: the unique job number for this simulation set
		:param int job_type: the job type (used for easy file identification after simulations are complete)
		:param str output_directory: the output directory to store the SQL database
		:param float min_speciation_rate: the minimum speciation rate to simulate
		:param float sigma: the dispersal sigma value
		:param float tau: the fat-tailed dispersal tau value
		:param int deme: the deme size (in individuals per cell)
		:param float sample_size: the sample size of the deme (decimal 0-1)
		:param float max_time: the maximum allowed simulation time (in seconds)
		:param str dispersal_method: the dispersal kernel method. Should be one of [normal, fat-tail, norm-uniform]
		:param float m_prob: the probability of drawing from the uniform dispersal. Only relevant for uniform dispersals
		:param float cutoff: the maximum value for the uniform dispersal. Only relevant for uniform dispersals.
		:param float dispersal_relative_cost: the relative cost of travelling through non-habitat (defaults to 1)
		:param int min_num_species: the minimum number of species known to exist (defaults to 1
		:param float habitat_change_rate: the rate of habitat change over time
		:param float gen_since_historical: the time in generations since a historical state was achieved
		:param bool restrict_self: if true, restricts dispersal from own cell
		:param bool/str landscape_type: if false or "closed", restricts dispersal to the provided maps, otherwise
										can be "infinite", or a tiled landscape using "tiled_coarse" or "tiled_fine".
		:param bool protracted: if true, uses protracted speciation application
		:param float min_speciation_gen: the minimum amount of time a lineage must exist before speciation occurs.
		:param float max_speciation_gen: the maximum amount of time a lineage can exist before speciating.
		:param bool spatial: if true, means that the simulation is spatial
		:param bool uses_spatial_sampling: if true, the sample mask is interpreted as a proportional sampling mask,
										   where the number of individuals sampled in the cell is equal to the
										   density * deme_sample * cell sampling proportion
		:param list times: list of temporal sampling points to apply (in generations)
		"""
		if not self.is_setup_param:
			self.seed = seed
			self.output_directory = output_directory
			self.job_type = job_type
			self.min_speciation_rate = min_speciation_rate
			self.sigma = sigma
			self.tau = tau
			self.deme = deme
			self.sample_size = sample_size
			self.max_time = max_time
			self.dispersal_relative_cost = dispersal_relative_cost
			self.min_num_species = min_num_species
			self.habitat_change_rate = habitat_change_rate
			self.gen_since_historical = gen_since_historical
			self.dispersal_method = dispersal_method
			self.m_prob = m_prob
			self.cutoff = cutoff
			self.calculate_sql_database()
			self.is_setup_param = True
			self.restrict_self = restrict_self
			self.is_spatial = spatial
			if uses_spatial_sampling and not spatial:
				raise ValueError("Must use a spatial simulation to define spatial samplign regime.")
			self.uses_spatial_sampling = uses_spatial_sampling
			if not spatial:
				self.is_setup_map = True
			if landscape_type in {False, "closed"}:
				self.landscape_type = "closed"
			elif landscape_type in {True, "infinite"}:
				self.landscape_type = "infinite"
			elif landscape_type in {"tiled_coarse", "tiled_fine"}:
				self.landscape_type = landscape_type
			else:
				raise ValueError("Supplied landscape type is not recognised: {}".format(landscape_type))
			if protracted:
				self.protracted = True
				if max_speciation_gen is None:
					self.logger.warning("Using protracted speciation, but no maximum speciation generation supplied."
										" Default to 10^100")
					self.max_speciation_gen = 10 ** 100
				else:
					self.max_speciation_gen = max_speciation_gen
				if min_speciation_gen is None:
					self.logger.warning("Using protracted speciation, but no minimum speciation generation supplied."
										" Default to 0.0")
					self.min_speciation_gen = 0.0
				else:
					self.min_speciation_gen = min_speciation_gen
			if times:
				self.times = times
		else:
			self.logger.warning("Parameters already set up.")

	def check_simulation_params(self):
		"""
		Checks that simulation parameters have been correctly set and the program is ready for running.
		Note that these checks have not been fully tested and are probably unnecessary in a large number of cases.
		"""
		if not self.is_setup_param:
			raise RuntimeError("Simulation parameters have not been set.")
		if self.output_directory in {"", None, 'null'}:
			raise RuntimeError('Output directory not set.')
		check_parent(self.output_directory)
		if self.full_config_file is not None:
			check_parent(self.full_config_file)

	def resume_coalescence(self, pause_directory, seed, job_type, max_time, out_directory=None,
						   protracted=None, spatial=None):
		"""
		Resumes the simulation from the specified directory, looking for the simulation with the specified seed and task
		referencing.

		:param pause_directory: the directory to search for the paused simulation
		:param seed: the seed of the paused simulation
		:param job_type: the task of the paused simulation
		:param max_time: the maximum time to run simulations for
		:param out_directory: optionally provide an alternative output location. Defaults to same location as
		pause_directory
		:param bool protracted: protractedness of the simulation
		:param bool spatial: if the simulation is to be run with spatial complexity

		.. note: A max time of 0 uses the previous simulation maximum time.

		:return: None
		"""
		if protracted is not None:
			self.protracted = protracted
		if spatial is not None:
			self.is_spatial = spatial
		if out_directory is None:
			out_directory = pause_directory
		self.output_directory = out_directory
		self.pause_directory = pause_directory
		self.seed = seed
		self.job_type = job_type
		self.max_time = max_time
		self.calculate_sql_database()
		file_path = [
			os.path.join(pause_directory, "Pause", str("Dump_" + x + "_" + str(job_type) + "_" + str(seed) + ".csv"))
			for x in
			["active", "data", "map"]]
		for each in file_path:
			if not os.path.exists(each):
				raise IOError(
					"Paused file " + each + " not found. Ensure pause directory is correct and is accessible.")
		self.setupNECSim()
		self.c_simulation.setup_resume(pause_directory, out_directory, seed, job_type, max_time)
		self.is_setup_complete = True
		self._run_and_output()

	def persistent_ram_usage(self):
		"""
		This is the persistent RAM usage which cannot be optimised by the program for a particular set of maps
		:return: the total persistent RAM usage in bytes
		"""
		total_ram = 0
		total_ram += self.fine_map.x_size * self.fine_map.y_size * 4
		# First add the maps
		if self.coarse_map.file_name not in [None, "none"]:
			total_ram += self.coarse_map.x_size * self.coarse_map.y_size * 4
		if self.historical_fine_map_file not in [None, "none"] or len(self.historical_fine_list) != 0:
			total_ram += self.fine_map.x_size * self.fine_map.y_size * 4
		if self.historical_coarse_map_file not in [None, "none"] or len(self.historical_coarse_list) != 0:
			total_ram += self.coarse_map.x_size * self.coarse_map.y_size * 4
		if self.reproduction_map.file_name not in [None, "none"]:
			total_ram += self.reproduction_map.x_size * self.reproduction_map.y_size * 8
		if self.sample_map.file_name not in [None, "none", "null"]:
			total_ram += self.sample_map.x_size * self.sample_map.y_size * 1
		if self.dispersal_map.file_name not in [None, "none"]:
			total_ram += self.fine_map.x_size * self.fine_map.y_size * 8
		# Now add the other large memory objects
		# The data object
		total_ram += 2 * self.count_individuals() * 91 * min(len(self.times_list), 1)
		# The active object
		total_ram += 72 * self.count_individuals() * min(len(self.times_list), 1)
		return total_ram

	def estimate_ram_usage(self, grid_individuals):
		"""
		Estimates the RAM usage for the program with the current parameters.

		Note this estimates the upper bound of memory usage and is likely inaccurate, especially for simulations with
		multiple time sampling points.

		:param grid_individuals: the number of individuals existing on the grid

		:return: the estimated RAM usage in bytes

		:rtype: float
		"""
		total_ram = self.persistent_ram_usage()
		# The grid object
		if self.grid.file_name not in [None, "none", "null"]:
			total_ram += (13 * (self.grid.x_size + self.grid.y_size)) + 8 * grid_individuals

	def count_individuals(self):
		"""
		Estimates the number of individuals to be simulated. This may be inaccurate if using multiple time points and
		historical maps.

		:return: a count of the number of individuals to be simulated

		:rtype: float
		"""
		if self.count_total is None:
			if self.sample_map.file_name in [None, "none", "null"]:
				ds = gdal.Open(self.fine_map.file_name)
			else:
				ds = gdal.Open(self.sample_map.file_name)
			self.count_total = np.sum(ds.GetRasterBand(1).ReadAsArray()) * \
							   self.sample_size * self.get_average_density()
			if len(self.times_list) > 0:
				self.count_total *= len(self.times_list)
			return self.count_total
		else:
			return self.count_total

	def import_fine_map_array(self):
		"""
		Imports the fine map array to the in-memory object, subsetted to the same size as the sample grid.

		:rtype: None
		"""
		if self.fine_map_array is None:
			if self.fine_map.file_name == "null":
				self.fine_map_array = np.ones((self.fine_map.x_size, self.fine_map.y_size))
			else:
				self.fine_map_array = self.fine_map.get_subset(self.fine_map.x_offset, self.fine_map.y_offset,
															   self.sample_map.x_size, self.sample_map.y_size)

	def import_sample_map_array(self):
		"""
		Imports the sample map array to the in-memory object.

		:rtype: None
		"""
		if self.sample_map_array is None:
			if self.sample_map.file_name == "null":
				self.sample_map_array = np.ones((self.sample_map.x_size, self.sample_map.y_size))
			else:
				self.sample_map.open()
				if not self.uses_spatial_sampling:
					self.sample_map_array = np.ma.masked_where(self.sample_map.data >= 0.5, self.sample_map.data).mask
				else:
					self.sample_map_array = self.sample_map.data
				self.sample_map.data = None

	def grid_density_estimate(self, x_off, y_off, x_dim, y_dim):
		"""
		Counts the density total for a subset of the grid by sampling from the fine map

		:note: This is an approximation (based on the average density of the fine map) and does not produce
		       a perfect value. This is done for performance reasons. The actual value can be obtained with
		       grid_density_actual().

		:param x_off: the x offset of the grid map subset
		:param y_off: the y offset of the grid map subset
		:param x_dim: the x dimension of the grid map subset
		:param y_dim: the y dimension of the grid map subset

		:return: an estimate of the total individuals that exist in the subset.

		:rtype: int
		"""
		self.import_fine_map_array()
		if self.sample_map.file_name not in ["none", "null", None] and \
				self.sample_map.file_name != self.fine_map.file_name:
			self.import_sample_map_array()
			# Less accurate, but faster way.
			arr_subset = self.sample_map_array[y_off:y_off + y_dim, x_off:x_off + x_dim]
			return int(np.sum(np.floor(arr_subset * self.get_average_density() * self.sample_size)))
		return int(np.sum(np.floor(self.fine_map_array * self.deme * self.sample_size)))

	def grid_density_actual(self, x_off, y_off, x_dim, y_dim):
		"""
		Counts the density total for a subset of the grid by sampling from the fine map.

		Note that for large maps this can take a very long time.

		:param x_off: the x offset of the grid map subset
		:param y_off: the y offset of the grid map subset
		:param x_dim: the x dimension of the grid map subset
		:param y_dim: the y dimension of the grid map subset

		:return: the total individuals that exist in the subset.

		:rtype: int
		"""
		self.import_fine_map_array()
		if self.sample_map.file_name not in ["none", "null", None] and \
				self.sample_map.file_name != self.fine_map.file_name:
			self.import_sample_map_array()
			# Less accurate, but faster way.
			return int(np.sum(np.floor(np.multiply(self.fine_map_array[y_off:y_off + y_dim, x_off:x_off + x_dim],
												   self.sample_map_array[y_off:y_off + y_dim, x_off:x_off + x_dim]) *
									   self.deme * self.sample_size)))
		else:
			return int(np.sum(np.floor(self.fine_map_array[y_off:y_off + y_dim, x_off:x_off + x_dim] *
									   self.deme * self.sample_size)))

	def get_average_density(self):
		"""
		Gets the average density across the fine map, subsetted for the sample grid.
		"""
		if self.fine_map_average is None:
			self.import_fine_map_array()
			self.fine_map_average = self.deme * np.mean(self.fine_map_array)
		return self.fine_map_average

	def check_sample_map_equals_sample_grid(self):
		"""
		Checks if the grid and sample map are the same size and offset (in which case, future operations can be
		simplified).

		:return: true if the grid and sample map dimensions and offsets are equal
		:rtype: bool
		"""
		return self.grid.x_size == self.sample_map.x_size and self.grid.y_size == self.sample_map.x_size and \
			   self.grid.x_offset == 0 and self.grid.y_offset == 0

	def optimise_ram(self, ram_limit):
		"""
		Optimises the maps for a specific RAM usage.

		If ram_limit is None, this function does nothing.

		:note: Assumes that the c++ compiler has sizeof(long) = 8 bytes for calculating space usage.

		:note: Only optimises RAM for a square area of the map. For rectangular shapes, will use the shortest length as
			   a maximum size.

		:param ram_limit: the desired amount of RAM to limit to, in GB

		:raises MemoryError: if the desired simulation cannot be compressed into available RAM
		"""
		if self.sample_size <= 0 or self.deme <= 0:
			raise ValueError("Sample size is 0, or deme is 0. set_simulation_params() before optimising RAM.")
		self.grid = copy.deepcopy(self.sample_map)
		# Over-estimate static usage slightly
		static_usage = 1.1 * self.persistent_ram_usage() / 1024 ** 3
		if static_usage > ram_limit:
			raise MemoryError(
				"Cannot achieve RAM limit: minimum requirements are {}GB.".format(round(static_usage, 2)))
		remaining_space = (ram_limit - static_usage) * 1024 ** 3
		self.logger.info("Remaining space is {}GB.\n".format(round(remaining_space / 1024 ** 3, 2)))
		# Rough calculation of number of cells we have remaining (aim for 75% to be sure)
		average_density = self.get_average_density()
		# This is the size of the grid object in memory
		tmp_space = 0.95 * (remaining_space / ((16 * average_density) + 32))
		size = int(np.rint(tmp_space ** 0.5))
		if size < self.sample_map.x_size or size < self.sample_map.y_size:
			if size > self.sample_map.x_size or size > self.sample_map.y_size:
				size = min(self.sample_map.x_size, self.sample_map.y_size)
			if size < 2:
				raise MemoryError("Could not find square grid to achieve RAM limit. Please set this manually.")
			max_attained = 0
			max_x = 0
			max_y = 0
			# Now loop over every square of the required size on the grid, until we find one with the maximum density
			if (self.sample_map.x_size * self.sample_map.y_size) > 1000000:
				printing = True
			else:
				printing = False
			end_x = self.sample_map.x_size - size - 1
			if size < 10:
				self.logger.warning("Extremely small grid size: {}. Disabling maximum density checking.".format(size))
				max_x = int(self.sample_map.x_size / 2)
				max_y = int(self.sample_map.y_size / 2)
				max_attained = 1
			else:
				for x in range(0, max(self.sample_map.x_size - size - 1, 1), int(round(max(10.0, size / 10)))):
					if printing:
						self.logger.info("\rChecking {} / {} : {}%".format(x, end_x, round(100 * x / end_x)))
					for y in range(0, max(self.sample_map.y_size - size - 1, 1), int(round(max(10.0, size / 10)))):
						# print("x, y: {}, {}".format(x, y))
						tmp_total = self.grid_density_actual(x, y, size, size)
						if tmp_total > max_attained:
							max_x = x
							max_y = y
							max_attained = tmp_total
				if printing:
					self.logger.info("\rChecking complete               \n")
			if max_attained > 0:
				# Now do one last check for consistency
				while self._fine_map_check(max_x, max_y, size):
					if printing:
						printing = False
						self.logger.info("Shrinking due to high densities in sample zone.\n")
					# Brute-force dividing by 2 each time
					size = max(1, int(size / 2))
				if max_x + size > self.sample_map.x_size or max_y + size > self.sample_map.y_size:
					self.logger.debug("Max x,y: {}, {}".format(max_x, max_y))
					self.logger.debug("Size: {}".format(size))
					self.logger.debug("Fine map size: {}, {}".format(self.fine_map.x_size,
																	 self.fine_map.y_size))
					self.logger.debug("Sample map size: {}, {}".format(self.sample_map.x_size,
																	   self.sample_map.y_size))
					raise RuntimeError("Incorrect dimension setting - please report this bug")
				self.grid.x_size = size
				self.grid.y_size = size
				self.sample_map.x_offset = max_x
				self.sample_map.y_offset = max_y
				self.grid.file_name = "set"
				if self.check_sample_map_equals_sample_grid():
					self.grid.file_name = "none"
			else:
				raise MemoryError("Could not optimise maps for desired memory usage. "
								  "Attempted grid size of {} and achived max of {}".format(size, max_attained))
		self._wipe_objects()

	def get_optimised_solution(self):
		"""
		Gets the optimised solution as a dictionary containing the important optimised variables.
		This can be read back in with set_optimised_solution

		:return: dict containing the important optimised variables
		:rtype: dict
		"""
		return {"grid_x_size": self.grid.x_size,
				"grid_y_size": self.grid.y_size,
				"sample_x_offset": self.sample_map.x_offset,
				"sample_y_offset": self.sample_map.y_offset,
				"grid_file_name": self.grid.file_name}.copy()

	def set_optimised_solution(self, dict_in):
		"""
		Sets the optimised RAM solution from the variables in the provided dictionary.
		This should contain the grid_x_size, grid_y_size, grid_file_name, sample_x_offset and sample_y_offset.

		:param dict dict_in: the dictionary containing the optimised RAM solution variables
		:rtype: None
		"""
		self.grid.x_size = dict_in["grid_x_size"]
		self.grid.y_size = dict_in["grid_y_size"]
		self.sample_map.x_offset = dict_in["sample_x_offset"]
		self.sample_map.y_offset = dict_in["sample_y_offset"]
		self.grid.file_name = dict_in["grid_file_name"]

	def finalise_setup(self, expected=False, ignore_errors=False):
		"""
		Runs all setup routines to provide a complete simulation. Should be called immediately before run_coalescence()
		to ensure the simulation setup is complete.

		:param ignore_errors: if true, any FileNotFoundError and FileExistsError raised by checking the output database
							  are ignored
		:param expected: set to true if we expect the output file to exist

		"""
		self.check_simulation_params()
		try:
			self.run_checks(expected=expected)
		except (FileExistsError, FileNotFoundError) as err:
			if not ignore_errors:
				raise err
			else:
				self.logger.info(str(err))
		# Now check if config files need to be written.
		if self.historical_fine_map_file not in self.historical_fine_list and \
				self.historical_fine_map_file not in [None, "", "none"]:
			self.add_historical_map(self.historical_fine_map_file, self.historical_coarse_map_file,
									self.gen_since_historical, self.habitat_change_rate)
		self.create_config()
		self.create_map_config()
		self.calculate_sql_database()
		self.setupNECSim()
		if self.full_config_file is not None:
			self.c_simulation.import_from_config(self.full_config_file)
		else:
			self.config.write(self.config_string)
			self.c_simulation.import_from_config_string(self.config_string.getvalue())
		self.c_simulation.setup()
		self.is_setup_complete = True

	def setupNECSim(self):
		"""
		Calculates the type of the simulation (spatial/non-spatial, protracted/non-protracted) and sets the c object
		appropriately.

		:rtype: None
		"""
		if self.protracted:
			if self.is_spatial:
				self.c_simulation = libnecsim.CPSpatialSimulation(self.logger, write_to_log)
			else:
				self.c_simulation = libnecsim.CPNSESimulation(self.logger, write_to_log)
		elif self.is_spatial:
			self.c_simulation = libnecsim.CSpatialSimulation(self.logger, write_to_log)
		else:
			self.c_simulation = libnecsim.CNSESimulation(self.logger, write_to_log)

	def set_speciation_rates(self, speciation_rates):
		"""Add speciation rates for analysis at the end of the simulation. This is optional

		:param list speciation_rates: a list of speciation rates to apply at the end of the simulation
		"""
		self.speciation_rates.extend(speciation_rates)

	def calculate_sql_database(self):
		"""
		Saves the output database location to self.output_database.
		"""
		self.output_database = os.path.join(self.output_directory, "data_{}_{}.db".format(str(self.job_type),
																						  str(self.seed)))

	def check_sql_database(self, expected=False):
		"""
		Checks whether the output database exists. If the existance does not match the expected variable, raises an
		error.

		:raises FileExistsError: if the file already exists when it's not expected to
		:raises FileNotExistsError: if the file does not exist when we expect it to

		:param expected: boolean for expected existance of the output file

		:rtype: None
		"""
		if not expected:
			if os.path.exists(self.output_database):
				try:
					t = CoalescenceTree(self.output_database)
				except IOError:
					pass
				else:
					raise FileExistsError("Output sql database already exists.")
		else:
			if not os.path.exists(self.output_database):
				raise FileNotFoundError("Output sql database {} not found.".format(self.output_database))

	def check_maps(self):
		"""
		Checks that the maps all exist and that the file structure makes sense.

		:raises TypeError: if a dispersal map or reproduction map is specified, we must have a fine map specified, but
			not a coarse map.

		:raises IOError: if one of the required maps does not exist

		:return: None
		"""
		if self.is_spatial:
			Landscape.check_maps(self)
			if self.grid.file_name == "set" and \
					self.sample_map.x_offset + self.grid.x_size > self.sample_map.x_size or \
					self.sample_map.y_offset + self.grid.y_size > self.sample_map.y_size:
				raise ValueError("Grid is not within the sample map - please check offsets of sample map.")
			# Now check our combination of dispersal map, reproduction map and infinite landscape with our fine/coarse maps
			# makes sense
			if self.dispersal_map.file_name not in {"none", None}:
				if self.coarse_map.file_name not in {None, "none"}:
					raise TypeError("Cannot use a coarse map if using a dispersal map. "
									"This feature is currently unsupported.")
			if self.reproduction_map.file_name not in {None, "none", "null"}:
				if self.coarse_map.file_name not in {None, "none", "null"}:
					raise TypeError("Cannot use a coarse map if using a reproduction map. "
									"This feature is currently unsupported.")

	def run_checks(self, expected=False):
		"""
		Check that the simulation is correctly set up and that all the required files exist.

		:param expected: set to true if we expect the output file to already exist

		:raises RuntimeError: if previous set-up routines are not complete
		"""
		if self.is_setup_map and self.is_setup_param:
			self.check_maps()
		else:
			err = "Set up is incomplete."
			raise RuntimeError(err)
		self.check_sql_database(expected=expected)

	def run(self):
		"""
		Convenience function which completes setp, runs the simulation and calculates the coalescence tree for the set
		speciation rates in one step.

		:rtype: None
		"""
		self.finalise_setup()
		self._run_and_output()

	def _run_and_output(self):
		"""
		Runs the simulation and outputs to the database. Should only be called internally.

		:return:
		"""
		if self.run_coalescence():
			self.apply_speciation_rates()
		else:
			self.logger.warning("Simulation did not complete in time specified.\n")
			self.logger.warning("Resume with extra time to continue.\n")

	def run_coalescence(self):
		"""
		Attempt to run the simulation with the given simulation set-up.
		This is the main routine performing the actual simulation which will take a considerable amount of time.

		:return: True if the simulation completes successfully, False if the simulation pauses.

		:rtype: bool
		"""
		if self.is_setup_complete:
			# Make the output directory if it doesn't yet exist
			if not os.path.exists(self.output_directory):
				os.makedirs(self.output_directory)
			# Call the c++ object and run the simulation
			return self.c_simulation.run()
		else:
			raise RuntimeError("Set up is incomplete.")

	def apply_speciation_rates(self, speciation_rates=None):
		"""
		Applies the speciation rates to the coalescence tree and outputs to the database.

		:param speciation_rates: a list of speciation rates to apply

		:rtype: None
		"""
		if speciation_rates:
			for each in speciation_rates:
				self.speciation_rates.append(each)
			self.speciation_rates = [x for x in set(self.speciation_rates)]
		self.c_simulation.apply_speciation_rates(self.speciation_rates)

	def _wipe_objects(self):
		"""
		Clears the arrays from memory
		:return:
		"""
		self.fine_map_array = None
		self.sample_map_array = None

	def _fine_map_check(self, max_x, max_y, size):
		"""
		Checks the total number of individuals in a subset from the fine map is less than the average density multiplied
		by the size of the square.

		:param max_x: the x value to start subsetting from
		:param max_y: the y value to start subsetting from
		:param size: the size of the square
		:return: bool true if it is smaller
		:rtype: bool
		"""
		self._fine_map_sum_res = self.grid_density_actual(max_x, max_y, size, size)
		return self._fine_map_sum_res > size * size * self.get_average_density()



"""
Generate the fragment config files from a supplied shapefile and a raster file to offset from.

The function :func:`generate_fragment_csv()` contains the full pipeline to generate the fragment csv.
"""
import csv

import os
try:
	from osgeo import gdal, ogr, osr
except ImportError as ie:
	try:
		import gdal
		from gdal import ogr, osr
	except ImportError:
		gdal = None
		raise ie

from .future_except import FileNotFoundError, FileExistsError
from .map import Map

def generate_fragment_csv(input_shapefile, input_raster, output_csv, field_name="fragment", field_area="area"):
	"""
	Generates the fragment csv from the provided shapefile and raster file. Coordinates for outputted to the csv are
	calculated from the extent of each polygon in the shapefile as their relative position on the input raster.

	The fragment extents are used solely, so overlapping extents of fragments results in individuals in those areas
	appearing in both fragments. Therefore, rectangular fragments alone should be used.

	.. important:: The input shapefile and raster must have the same projection.

	:param input_shapefile: the shapefile containing polygons defining fragments. Should contain fields of field_name
							and field_area
	:param input_raster: raster file to calculate the relative coordinates on
	:param output_csv: output csv to create
	:param field_name: name of the field in the shapefile to acquire fragment names from
	:param field_area: name of the field in the shapefile to acquire the number of individuals from
	"""
	f = FragmentConfigHandler()
	f.generate_config(input_shapefile, input_raster, field_name, field_area)
	f.write_csv(output_csv)

class FragmentConfigHandler(object):
	"""
	Contains routines for calculating the offsets from a config file.
	"""

	def __init__(self):
		self.fragment_list = {}

	def generate_config(self, input_shapefile, input_raster, field_name="fragment", field_area="area"):
		"""
		Generates the config file from the shapefile containing the fragments, writing the coordinates of the extent of
		each fragment to the output csv. The coordinates are calculated from their relevant position on the input
		raster.

		:param str input_shapefile: shapefile containing the fragments in a "fragments" field, with each defined as a
							    polygon.
		:param str input_raster: the raster to calculate the coordinates from
		:param str field_name: optionally provide a field to extract fragment names from
		:param str field_area: optionally provide a field to extract fragment areas from (the number of individuals that
							   exist in the fragment.
		"""
		if not input_shapefile.endswith(".shp"):
			raise ValueError("Input file {} is not a shape file (.shp).".format(input_shapefile))
		if not input_raster.endswith(".tif"):
			raise ValueError("Input raster {} is not a tif file (.tif).".format(input_raster))
		if not os.path.exists(input_shapefile):
			raise FileNotFoundError("Input shapefile {} does not exist.".format(input_shapefile))
		if not os.path.exists(input_raster):
			raise FileNotFoundError("input raster {} does not exist.".format(input_raster))
		m = Map(input_raster)
		dim_x, dim_y = m.get_x_y()
		# Read the input Layer
		src_driver = ogr.GetDriverByName("ESRI Shapefile")
		src_ds = src_driver.Open(input_shapefile, gdal.GA_ReadOnly)
		if src_ds is None:
			raise IOError("Could not open {}".format(input_shapefile))
		src_layer = src_ds.GetLayer()
		for feature in src_layer:
			geom = feature.GetGeometryRef()
			fragment = feature.GetField(field_name)
			area = feature.GetField(field_area)
			min_long, max_long, min_lat, max_lat = geom.GetEnvelope()
			min_x, min_y = m.convert_lat_long(max_lat, min_long)
			max_x, max_y = m.convert_lat_long(min_lat, max_long)
			if min_x < 0 or min_y < 0 or max_x < 0 or max_y < 0 or \
					min_x > dim_x or min_y > dim_y or max_x > dim_x or max_y > dim_y:
				src_ds = None
				raise ValueError("Shapefile is not contained within the raster - cannot generate fragment config.")
			self.fragment_list[fragment] =  {"min_x" : min_x, "max_x" : max_x, "min_y" : min_y, "max_y" : max_y,
											 "area" : area}
		src_ds = None

	def write_csv(self, output_csv):
		"""
		Writes the fragments to the output csv.

		:param str output_csv: the csv to write the output to

		"""
		if os.path.exists(output_csv):
			raise FileExistsError("Output csv {} already exists. Please remove first.".format(output_csv))
		with open(output_csv, "w") as csv_file:
			csv_writer = csv.writer(csv_file)
			for key, value in sorted(self.fragment_list.items()):
				csv_writer.writerow([key, value["min_x"], value["min_y"], value["max_x"], value["max_y"],
									 value["area"]])


"""When run from the command line, installs pycoalescence to the environment.

Run ```python setup.py build``` to build the c++ module.

Run ```python setup.py install``` to build the c++ module and install it to the environment.

Run ```python setup.py develop``` to install the package in development mode.
"""
import logging
import os

from setuptools import setup, Extension, find_packages

logging.basicConfig()
logging.getLogger().setLevel(logging.CRITICAL)
necsim_path = os.path.join("pycoalescence", "necsim", "libnecsim.so")
if os.path.exists(necsim_path):
	os.remove(necsim_path)
try:
	from pycoalescence import __version__ as p_version
	from pycoalescence.installer import Installer
except ImportError:
	# Means that one of the dependencies isn't installed properly - this is fine
	pass

# Get all the c++ source files
# py_src_files = [os.path.join("pycoalescence/lib", x) for x in os.listdir("pycoalescence/lib")
# 				if ".cpp" in x]
# necsim_src_files = [os.path.join("pycoalescence/lib/necsim", x) for x in os.listdir("pycoalescence/lib/necsim") if
# 					".cpp" in x]
# all_srcs = [x for x in necsim_src_files + py_src_files if "main.cpp" not in x]
#
# data_dir = "pycoalescence/reference"
#
# data_files = [os.path.join(data_dir, x) for x in os.listdir(data_dir) if x not in ['Makefile', '.DS_Store',
# 																				   'log_12_2.txt']]


class CustomExtension(Extension):
	def __init__(self, name, sourcedir=''):
		Extension.__init__(self, name, sources=[])
		self.sourcedir = os.path.abspath(sourcedir)


with open('README.rst') as f:
	readme = f.read()

# libnecsim = Extension('pycoalescence/lib',
# 					  sources=all_srcs,
# 					  include_dirs = ["pycoalescence/lib" "pycoalescence/lib/necsim"],
# 					  libraries=['gdal', 'boost_filesystem', 'boost_system', 'sqlite3'],
# 					  extra_compile_args=['-std=c++14'],
# 					  language="c++",  # generate C++ code
# 					  )

setup(name='pycoalescence',
	  version=p_version,
	  description='A spatially-explicit neutral ecology simulator using coalescence methods',
	  author='Sam Thompson',
	  author_email='samuel.thompson14@imperial.ac.uk',
	  url='http://pycoalescence.readthedocs.io/',
	  long_description=readme,
	  # ext_modules=[libnecsim],
	  ext_modules=[CustomExtension("libnecsim", os.path.join("pycoalescence","lib"))],
	  cmdclass=dict(build_ext=Installer),
	  license='MIT',
	  packages=find_packages(exclude=["*tests*", 'docs']),
	  package_data={
		  'pycoalescence': ['reference/*.json', 'reference/*.json'],
	  },
	  classifiers=['Development Status :: 4 - Beta',
				   'License :: OSI Approved :: MIT License',
				   'Operating System :: MacOS',
				   'Operating System :: MacOS :: MacOS X',
				   # 'Operating System :: Microsoft :: Windows :: Windows 10',
				   'Operating System :: POSIX',
				   'Programming Language :: C++',
				   'Programming Language :: Python :: 2.7',
				   'Programming Language :: Python :: 3.6',
				   'Programming Language :: Python :: 3',
				   'Programming Language :: Python :: Implementation :: CPython',
				   'Topic :: Scientific/Engineering',
				   'Intended Audience :: Science/Research',
				   'Natural Language :: English',
				   ],
	  zip_safe=False,
	  keywords='neutral simulation ecology spatially-explicit',
	  install_requires=['GDAL>=1.11.2', 'numpy'],
	  extras_require={
		  'scipy': ['scipy>=0.12.0'],
		  'plotting': ['matplotlib']
	  }
	  )
